/*
   To reconstruct an ELF executable binary from

   a raw Linux image and its corresponding symbol map file

 */

#define _GNU_SOURCE
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "compat.h"
#include "libelf.h"
#include "gelf.h"
#include "native-elf-format.h"

#define SEPARATOR '\0'

char *outputFileName = NULL;
char deleteOutputFile = 1;

#if defined(__USE_GNU) || defined(__USE_XOPEN2K8)
#define createTempFile(fileHandle) \
    do { \
        char *buf = NULL; \
        size_t bufSize = 0; \
        fileHandle = open_memstream(&buf, &bufSize); \
        if (NULL == fileHandle) \
            err ( EX_SOFTWARE , "Cannot create temporary file using open_memstrem() at %s:%d", __FILE__, __LINE__); \
    } while (0)
#else
#define createTempFile(fileHandle) \
    do { \
        fileHandle = tmpfile(); \
        if (NULL == fileHandle) \
            err ( EX_SOFTWARE , "Cannot create temporary file using tmpfile() at %s:%d", __FILE__, __LINE__); \
    } while (0)
#endif

void loadSymbolMapFile( FILE *symbolMapFile, FILE **symtabSection, FILE **strtabSection ) {
    char type, sym[512];
    unsigned long int address;
    int lineRead = 0;
#if LIBELF_CLASS == ELFCLASS64
    Elf64_Sym elfSymbol;
#elif LIBELF_CLASS == ELFCLASS32
    Elf32_Sym elfSymbol;
#endif

    /* create temporary files to store .symtab and .strtab sections */
    createTempFile( *symtabSection );
    createTempFile( *strtabSection );

    rewind( symbolMapFile );
    memset( &elfSymbol, 0, sizeof( elfSymbol ) );

    /*
        the idea of how to parse the symbol map file is from sysklogd's ksym.c

        Sysklogd is available at http://freshmeat.net/projects/sysklogd/

     */

    while ( !feof( symbolMapFile ) ) {
        ++lineRead;
        if ( 3 != fscanf( symbolMapFile, "%lx %c %511s\n", &address, &type, sym ) )
            errx ( EX_SOFTWARE , "The symbol map file is malformed at line %d", lineRead );
        elfSymbol.st_value = address;
        elfSymbol.st_info  = type;  /* we will correct st_info later */
        elfSymbol.st_name  = ftell( *strtabSection );

        if ( 1 != fwrite( &elfSymbol, sizeof( elfSymbol ), 1, *symtabSection ) )
            err ( EX_SOFTWARE , "Cannot write to temporary file for .symtab section when processing line %d", lineRead );

        if ( 0 >= fprintf( *strtabSection, "%s%c" , sym, SEPARATOR ) )
            err ( EX_SOFTWARE , "Cannot write to temporary file for .strtab section when processing line %d", lineRead );
    }
}

static Elf_Scn *createSection( Elf *elf,
                               unsigned long secType,
                               unsigned long secFlag,
                               unsigned long secAddr,
                               const char *secName,
                               size_t *num,
                               FILE *shstrtabSection ) {
    assert( NULL != elf );
    assert( NULL != shstrtabSection );
    Elf_Scn *scn;

    if ( ( scn = elf_newscn( elf ) ) == NULL )
        errx ( EX_SOFTWARE , "Cannot create a new section at %s:%d", __FILE__, __LINE__ );

    {
#if LIBELF_CLASS == ELFCLASS64
        Elf64_Shdr *shdr = elf64_getshdr( scn );
#elif LIBELF_CLASS == ELFCLASS32
        Elf32_Shdr *shdr = elf32_getshdr( scn );
#endif
        if ( NULL == shdr )
            errx ( EX_SOFTWARE , "Cannot get section header at %s:%d", __FILE__, __LINE__ );
        shdr->sh_type = secType;
        shdr->sh_flags = secFlag;
        shdr->sh_addr = secAddr;
        shdr->sh_name = ftell( shstrtabSection );
        shdr->sh_entsize = 0;
        if ( NULL == num || 0 == *num ) {
            if ( 0 >= fprintf( shstrtabSection, "%s%c", secName, SEPARATOR ) )
                err ( EX_SOFTWARE , "Cannot write to temporary file for .shstrtab section when creating section %s", secName );
            if ( NULL != num ) ++*num;
        }
        else {
            if ( 0 >= fprintf( shstrtabSection, "%s%zu%c", secName, 1 + *num, SEPARATOR ) )
                err ( EX_SOFTWARE , "Cannot write to temporary file for .shstrtab section when creating section %s", secName );
            ++*num;
        }
    }
    return scn;
}

static void setElfHeader( Elf *elf ) {
#if LIBELF_CLASS == ELFCLASS64
    Elf64_Ehdr *ehdr = elf64_newehdr( elf );
#elif LIBELF_CLASS == ELFCLASS32
    Elf32_Ehdr *ehdr = elf32_newehdr( elf );
#endif

    if ( NULL == ehdr )
        errx ( EX_SOFTWARE , "Cannot set ELF header at %s:%d", __FILE__, __LINE__ );

    ehdr->e_ident[EI_DATA] = ELFTC_BYTEORDER;
    ehdr->e_machine = ELFTC_ARCH;
    ehdr->e_type = ET_EXEC;
}

static void cleanUp() {
#if defined(__USE_GNU)
    fcloseall();
#endif
    if ( deleteOutputFile && outputFileName )
        unlink( outputFileName );
}

static int addressComp( const void *a, const void *b ) {
#if LIBELF_CLASS == ELFCLASS64
    Elf64_Sym *sa = ( Elf64_Sym * )a;
    Elf64_Sym *sb = ( Elf64_Sym * )b;
#elif LIBELF_CLASS == ELFCLASS32
    Elf32_Sym *sa = ( Elf32_Sym * )a;
    Elf32_Sym *sb = ( Elf32_Sym * )b;
#endif

    if ( sa->st_value == sb->st_value ) return 0;
    else {
        if ( sa->st_value < sb->st_value )
            return -1;
    }
    return 1;
}

#define PROCESS_SYMBOL(SECTYPE, SECNAME, SECNAME_COUNTER, SH_TYPE, SH_FLAGS, ST_TYPE) \
    do { \
        if (SECTYPE != currentSecType && 1024*1024*50 > elfSymbols[i].st_value - baseAddr) { \
            Elf_Scn *scn = createSection(elf, SH_TYPE, SH_FLAGS, elfSymbols[i].st_value, SECNAME, &SECNAME_COUNTER, shstrtabSection); \
            Elf_Data *data = elf_newdata(scn); \
            if (data == NULL) errx ( EX_SOFTWARE , "elf_newdata failed for " SECNAME "section at %s:%d", __FILE__, __LINE__); \
            data->d_align = sizeof(void*); \
            data->d_type  = ELF_T_BYTE; \
            data->d_off   = 0LL; \
            data->d_version = EV_CURRENT; \
            data->d_buf = rawImage+(elfSymbols[i].st_value - baseAddr); \
            if (elfSymbols[i].st_value % data->d_align) data->d_align = 1; \
            data->d_size = lastAddr - elfSymbols[i].st_value; \
            if (prevdata) prevdata->d_size = data->d_buf - prevdata->d_buf; \
            prevdata = data; \
            currentSecType = SECTYPE; \
            currentSecNdx = elf_ndxscn(scn); \
        } \
        elfSymbols[i].st_shndx = currentSecNdx; \
        elfSymbols[i].st_info  = GELF_ST_INFO(isLocal,ST_TYPE); \
        if (numSymbols != i) elfSymbols[i].st_size = elfSymbols[i+1].st_value - elfSymbols[i].st_value; \
    } while (0)

int main ( int argc , char **argv ) {
    FILE *rawImageFile, *symbolMapFile, *outputFile, *outputFile2;
    FILE *symtabSection, *strtabSection, *shstrtabSection;
    size_t numBSS = 0;
    size_t numData = 0;
    size_t numText = 0;
    size_t numReadOnly = 0;
    size_t numSymbols = 0;
    int i;
    char currentSecType = 0;
    size_t currentSecNdx;
#if LIBELF_CLASS == ELFCLASS64
    Elf64_Sym *elfSymbols = NULL, *elfSymbol;
#elif LIBELF_CLASS == ELFCLASS32
    Elf32_Sym *elfSymbols = NULL, *elfSymbol;
#endif

    Elf *elf;
    Elf_Data *prevdata = NULL;
    typeof( elfSymbols->st_value ) baseAddr, lastAddr;
    char *rawImage;

    if ( argc < 4 )
        errx ( EX_USAGE , "Usage: %s rawImageFile symbolMapFile outputFile ", argv [0] );

    if ( elf_version ( EV_CURRENT ) == EV_NONE )
        errx ( EX_SOFTWARE , "elf_version failed : %s", elf_errmsg ( -1 ) );

    /* open the raw image for reading */
    if ( ( rawImageFile = fopen ( argv [1] , "r" ) ) == NULL )
        err ( EX_NOINPUT , "Cannot open raw Linux kernel image file %s for reading", argv [1] );

    /* check if it's already an ELF executable binary */
    if ( ( elf = elf_begin( fileno( rawImageFile ), ELF_C_READ, NULL ) ) == NULL )
        err ( EX_NOINPUT , "Cannot determine the type of %s", argv [1] );

    if ( ELF_K_ELF == elf_kind( elf ) )
        errx ( EX_NOINPUT , "%s is already an ELF executable binary", argv [1] );

    if ( ELF_K_NONE != elf_kind( elf ) )
        errx ( EX_NOINPUT , "%s does not seem to be a raw Linux kernel image", argv [1] );
    elf_end( elf );

    /* open the symbol map file for reading */
    if ( ( symbolMapFile = fopen ( argv[2], "r" ) ) == NULL )
        err ( EX_NOINPUT , "Cannot open symbol map file %s for reading", argv [2] );

    /* load the symbol map file */
    loadSymbolMapFile( symbolMapFile, &symtabSection, &strtabSection );
    i = ftell( symtabSection );
    assert( 0 < i );
    numSymbols = i / sizeof( *elfSymbols );

    elfSymbols = malloc( i );
    assert( elfSymbols );

    /* load the .symtab temporary file */
    rewind( symtabSection );
    if ( numSymbols != fread( elfSymbols, sizeof( *elfSymbols ), numSymbols, symtabSection ) )
        err ( EX_SOFTWARE , "Cannot read from temporary file for .symtab section" );
    fclose( symtabSection ); /* possible memory leak here if we use open_memstream */

    /* sort the symbol map according to the symbol addresses */
    qsort( elfSymbols, numSymbols, sizeof( *elfSymbols ), addressComp );

    /* open the output file for writing */
    if ( ( outputFile = fopen ( argv [3] , "w" ) ) == NULL )
        err ( EX_NOINPUT , "Cannot open output file %s", argv [3] );
    outputFileName = argv[3];
    atexit( cleanUp ); /* delete output file if there is any error */
    if ( ( elf = elf_begin( fileno( outputFile ), ELF_C_WRITE, NULL ) ) == NULL )
        err ( EX_SOFTWARE , "elf_begin failed on the temporary output file at %s:%d", __FILE__, __LINE__ );
    setElfHeader( elf );
    createTempFile( shstrtabSection ); /* create temporary files to store .shstrtab section */
    fprintf( shstrtabSection, "%c", SEPARATOR ); /* first section is a dummy section with no name */

    /* map in the raw image */
    rewind( rawImageFile );
    {
        struct stat stat_buf;
        fstat( fileno( rawImageFile ), &stat_buf );
        lastAddr = stat_buf.st_size;
        rawImage = ( char * )mmap ( 0, stat_buf.st_size, PROT_READ, MAP_PRIVATE, fileno( rawImageFile ), 0 );
    }
    if ( MAP_FAILED == rawImage ) err ( EX_NOINPUT , "Cannot map in the raw Linux kernel image file" );
    fclose( rawImageFile );

    /* go through the symbols */
    baseAddr = elfSymbols[0].st_value;
    lastAddr += baseAddr;
    for ( i = 0; i < numSymbols; ++i ) {
        typeof( elfSymbols->st_info ) isLocal = ( 'a' <= elfSymbols[i].st_info && 'z' >= elfSymbols[i].st_info ) ? STB_LOCAL : STB_GLOBAL;

        switch ( toupper( elfSymbols[i].st_info ) ) {
        case 'B':
            PROCESS_SYMBOL( 'B', ".bss", numBSS, SHT_NOBITS, SHF_ALLOC | SHF_WRITE, STT_OBJECT );
            break;
        case 'D':
            PROCESS_SYMBOL( 'D', ".data", numData, SHT_PROGBITS, SHF_ALLOC | SHF_WRITE, STT_OBJECT );
            break;
        case 'R':
            PROCESS_SYMBOL( 'R', ".rodata", numReadOnly, SHT_PROGBITS, SHF_ALLOC, STT_OBJECT );
            break;
        case 'T':
            PROCESS_SYMBOL( 'T', ".text", numText, SHT_PROGBITS, SHF_ALLOC | SHF_EXECINSTR, STT_FUNC );
            break;
        case 'A':
            elfSymbols[i].st_shndx = SHN_ABS;
            elfSymbols[i].st_info  = GELF_ST_INFO( isLocal, STT_NOTYPE );
            break;
        case 'W':
            elfSymbols[i].st_shndx = currentSecNdx;
            elfSymbols[i].st_info  = GELF_ST_INFO( STB_WEAK, STT_FUNC );
            break;
        default:
            errx ( EX_SOFTWARE , "Cannot handle symbol type '%c' at line %d", elfSymbols[i].st_info, 1 + i );
        }
    }

    /* create .symtab section */
    {
        Elf_Scn *scn = createSection( elf, SHT_SYMTAB, 0, 0, ".symtab", NULL, shstrtabSection );
#if LIBELF_CLASS == ELFCLASS64
        Elf64_Shdr *shdr = elf64_getshdr( scn );
#elif LIBELF_CLASS == ELFCLASS32
        Elf32_Shdr *shdr = elf32_getshdr( scn );
#endif
        Elf_Data *data = elf_newdata( scn );
        shdr->sh_entsize = sizeof( *elfSymbols );
        shdr->sh_link = elf_ndxscn( scn ) + 1; /* the section index of .strtab section */
        data->d_type  = ELF_T_SYM;
        data->d_align = sizeof( void * );
        data->d_size  = numSymbols * sizeof( *elfSymbols );
        data->d_buf   = elfSymbols;
        data->d_off   = 0LL;
        data->d_version = EV_CURRENT;
    }

    /* create .strtab section */
    {
        Elf_Scn *scn = createSection( elf, SHT_STRTAB, SHF_STRINGS, 0, ".strtab", NULL, shstrtabSection );
        Elf_Data *data = elf_newdata( scn );
        char *t;
        data->d_type  = ELF_T_BYTE;
        data->d_align = 1;
        data->d_size  = ftell( strtabSection );
        rewind( strtabSection );
        data->d_buf = malloc( data->d_size );
        assert( NULL != data->d_buf );
        if ( data->d_size != fread( data->d_buf, 1, data->d_size, strtabSection ) )
            err ( EX_SOFTWARE , "Cannot read from temporary file for .strtab section" );
        fclose( strtabSection );
        data->d_off   = 0LL;
        data->d_version = EV_CURRENT;
    }

    /* create .shstrtab section */
    {
        Elf_Scn *scn = createSection( elf, SHT_STRTAB, SHF_STRINGS, 0, ".shstrtab", NULL, shstrtabSection );
        Elf_Data *data = elf_newdata( scn );
        char *t;
        data->d_type  = ELF_T_BYTE;
        data->d_align = 1;
        data->d_size  = ftell( shstrtabSection );
        rewind( shstrtabSection );
        data->d_buf = malloc( data->d_size );
        assert( NULL != data->d_buf );
        if ( data->d_size != fread( data->d_buf, 1, data->d_size, shstrtabSection ) )
            err ( EX_SOFTWARE , "Cannot read from temporary file for .shstrtab section" );
        fclose( shstrtabSection );
        data->d_off   = 0LL;
        data->d_version = EV_CURRENT;
        elf_setshstrndx( elf, elf_ndxscn( scn ) );
    }

    if ( 0 > elf_update( elf, ELF_C_NULL ) )
        err ( EX_SOFTWARE , "elf_update failed at %s:%d", __FILE__, __LINE__ );

    /* create one program header */
    {
        GElf_Ehdr ehdr;
#if LIBELF_CLASS == ELFCLASS64
        Elf64_Phdr *phdr = elf64_newphdr( elf, 1 );
#elif LIBELF_CLASS == ELFCLASS32
        Elf32_Phdr *phdr = elf32_newphdr( elf, 1 );
#endif
        if ( NULL == gelf_getehdr( elf, &ehdr ) )
            err ( EX_SOFTWARE , "gelf_getehdr failed at %s:%d", __FILE__, __LINE__ );

        ehdr.e_entry = baseAddr; /* set the Entry Point Address */
        gelf_update_ehdr( elf, &ehdr );
        phdr->p_filesz = gelf_fsize( elf, ELFTC_CLASS, 1, EV_CURRENT );
        phdr->p_type = PT_PHDR;
        phdr->p_offset = ehdr.e_phoff;
        elf_flagphdr( elf, ELF_C_SET, ELF_F_DIRTY );
    }

    if ( 0 > elf_update( elf, ELF_C_WRITE ) )
        err ( EX_SOFTWARE , "elf_update failed at %s:%d", __FILE__, __LINE__ );
    elf_end( elf );
    fclose( outputFile );
    deleteOutputFile = 0;
}